/*
* morse.c
*
*/

#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>

#include <linux/kernel.h>	/* printk() */
#include <linux/slab.h>		/* kmalloc() */
#include <linux/fs.h>		/* everything... */
#include <linux/errno.h>	/* error codes */
#include <linux/types.h>	/* size_t */
#include <linux/proc_fs.h>
#include <linux/cdev.h>

#include <asm/system.h>		/* cli(), *_flags */
#include <asm/uaccess.h>	/* copy_*_user */

#include "morse.h"		/* local definitions */

/*
 * Our parameters which can be set at load time.
 */

int morse_major =   MORSE_MAJOR;
int morse_minor =   0;
int morse_nr_devs = MORSE_NR_DEVS;	/* number of bare morse devices */


#define BUFFER_SIZE 4000 
MODULE_AUTHOR("Jonathan Piat");
MODULE_LICENSE("GPL");

struct morse_dev * morse_devices;	/* allocated in morse_init_module */



int init_circular_buffer(struct circular_buffer * c_buffer, int size){
	c_buffer -> buffer = kmalloc(size * sizeof(char), GFP_KERNEL);
	if(!c_buffer -> buffer){
			return -1 ;
	}
	memset(c_buffer -> buffer, 0, size * sizeof(char));
	c_buffer -> size = size ;
	c_buffer -> read_pos = 0 ;
	c_buffer -> write_pos = 0 ;
	return 0 ;
}

int write_circular_buffer(size_t count, struct circular_buffer * c_buffer, char * data){	
	int ret_val ;	
	int remains = count ;
	int index = 0;
	int global_index = 0 ;	
	ret_val = count ;
	while(remains > 0){
		char * code = NULL ;
		char val = data[index] ;		
		if(val > 96){
			val = val - 97;
		}else if(val > 64){
			val = val - 65;
		}
		if(val < 25 && val >= 0){
			code = codes[val];
		}else if(val == 32){
			code = "|";		
		}
		if(code != NULL){
			int char_index = 0 ;
			printk("code is %s\n", code);		
			while(code[char_index] != '\0' && char_index < 6){
				c_buffer -> buffer[c_buffer -> write_pos] = code[char_index] ;
				c_buffer -> write_pos = (c_buffer -> write_pos + 1) % c_buffer -> size ;
				char_index ++ ;	
				global_index += 1 ;
			}	
		}
		remains  -- ;
		index ++ ;
		c_buffer -> buffer[c_buffer -> write_pos]	= ' ';
		c_buffer -> write_pos = (c_buffer -> write_pos + 1) % c_buffer -> size ;
	}
	return ret_val ;
}

int read_circular_buffer(size_t count, struct circular_buffer * c_buffer, char * data){
	int ret_val = count ;	
	int read_index = c_buffer -> read_pos;
	int write_index = c_buffer -> write_pos;
	int max_read ;
	if(read_index == write_index){
		ret_val = 0 ;
	}else if(read_index < write_index){
		max_read = write_index - read_index ;
		ret_val = min((int)max_read, (int)count);
		memcpy(data, &(c_buffer -> buffer[read_index]), ret_val*sizeof(char));
		c_buffer -> read_pos += ret_val ;
	}else{
		int remains = 0 ;
		int buffer_pos = 0 ;
		max_read = ((c_buffer -> size) - read_index ) + write_index ;
		ret_val = min((int)max_read, (int)count);
		remains = ret_val ;
		if(remains > (c_buffer -> size - read_index)){
			memcpy(data, &(c_buffer -> buffer[read_index]), (c_buffer -> size - read_index)*sizeof(char));
			remains = remains - (c_buffer -> size - read_index);
			c_buffer -> read_pos = 0 ;
			buffer_pos = (c_buffer -> size - read_index);
		}
		memcpy(&data[buffer_pos], &(c_buffer -> buffer[read_index]), remains*sizeof(char));
		c_buffer -> read_pos += remains ;
	}
	return ret_val ;	
}



/*
 * Open and close
 */

int morse_open(struct inode *inode, struct file *filp)
{
	struct morse_dev *dev; /* device information */
	dev = container_of(inode->i_cdev, struct morse_dev, cdev);
	filp->private_data = dev; /* for other methods */
	return 0;          /* success */
}

int morse_release(struct inode *inode, struct file *filp)
{
	return 0;
}


/*
 * Data management: read and write
 */

ssize_t morse_read(struct file *filp, char __user *buf, size_t count,
                loff_t *f_pos)
{
	int ret_val ;
	char * read_data ;	
	struct morse_dev *dev = filp->private_data; 
	struct circular_buffer * c_buffer = dev -> data ;
	read_data = kmalloc(count * sizeof(char), GFP_KERNEL);	
	printk("Trying to read %d bytes\n", count);
	ret_val = read_circular_buffer(count, c_buffer, read_data);
	copy_to_user(buf, read_data, ret_val * sizeof(char));

	return ret_val;
}

ssize_t morse_write(struct file *filp, const char __user *buf, size_t count,
                loff_t *f_pos)
{
	struct morse_dev *dev = filp->private_data; 
	struct circular_buffer * c_buffer = dev -> data ;
	char * local_buffer = kmalloc(count*sizeof(char), GFP_KERNEL);;
	if(copy_from_user(local_buffer, buf, count*sizeof(char))){
			printk("Error in copy from user\n");	
			return -EFAULT;		
		}
	
	return write_circular_buffer(count, c_buffer, local_buffer);
}




struct file_operations morse_fops = {
	.owner =    THIS_MODULE,
	.read =     morse_read,
	.write =    morse_write,
	.open =     morse_open,
	.release =  morse_release,
};

/*
 * Finally, the module stuff
 */

/*
 * The cleanup function is used to handle initialization failures as well.
 * Thefore, it must be careful to work correctly even if some of the items
 * have not been initialized
 */
void morse_cleanup_module(void)
{
	int i;
	dev_t devno = MKDEV(morse_major, morse_minor);

	/* Get rid of our char dev entries */
	if (morse_devices) {
		for (i = 0; i < morse_nr_devs; i++) {
			kfree(&morse_devices[i].data -> buffer);
			cdev_del(&morse_devices[i].cdev);
		}
		kfree(morse_devices);
	}
	/* cleanup_module is never called if registering failed */
	unregister_chrdev_region(devno, morse_nr_devs);
}


/*
 * Set up the char_dev structure for this device.
 */
static void morse_setup_cdev(struct morse_dev *dev, int index)
{
	int err, devno = MKDEV(morse_major, morse_minor + index);
	cdev_init(&dev->cdev, &morse_fops);
	dev->cdev.owner = THIS_MODULE;
	dev->cdev.ops = &morse_fops;
	dev->data = kmalloc(sizeof(struct circular_buffer), GFP_KERNEL);
	err = init_circular_buffer(dev->data, BUFFER_SIZE);
	if (err)
		printk(KERN_NOTICE "Error %d adding morse%d", err, index);
	err = cdev_add (&dev->cdev, devno, 1);
	/* Fail gracefully if need be */
	if (err)
		printk(KERN_NOTICE "Error %d adding morse%d", err, index);
	printk("morse device on %d %d\n", morse_major, morse_minor + index);	
}


int morse_init_module(void)
{
	int result, i;
	dev_t dev = 0;

/*
 * Get a range of minor numbers to work with, asking for a dynamic
 * major unless directed otherwise at load time.
 */
	result = alloc_chrdev_region(&dev, morse_minor, morse_nr_devs,
			"morse");
	morse_major = MAJOR(dev);

	if (result < 0) {
		printk(KERN_WARNING "morse: can't get major %d\n", morse_major);
		return result;
	}


	morse_devices = kmalloc(morse_nr_devs * sizeof(struct morse_dev), GFP_KERNEL);
	if (!morse_devices) {
		result = -ENOMEM;
		goto fail;  /* Make this more graceful */
	}
	memset(morse_devices, 0, morse_nr_devs * sizeof(struct morse_dev));

        /* Initialize each device. */
	for (i = 0; i < morse_nr_devs; i++) {
		morse_setup_cdev(&morse_devices[i], i);
	}
	return 0; /* succeed */

  fail:
	morse_cleanup_module();
	return result;
}

module_init(morse_init_module);
module_exit(morse_cleanup_module);
